# -*- coding: utf-8 -*-
#
# Copyright (c) Panu Lahtinen
#
# Author(s):
#
#   Panu Lahtinen <pnuu+git@iki.fi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import defusedxml.ElementTree as ET
import datetime as dt

import numpy as np

from fmiopendata import wfs
from fmiopendata.utils import read_url

TIME_FORMAT = "%Y-%m-%dT%H:%M:%SZ"


class MultiPoint(object):
    """Class for holding multipoint data."""

    def __init__(self, xml, query_id, timeseries=False):
        """Initialize class."""
        self._xml = ET.fromstring(xml)
        self.data = dict()
        self.location_metadata = dict()
        self._location2name = dict()
        self._timeseries = timeseries

        if "radionuclide-activity-concentration" in query_id:
            self._parse_radionuclide()
        else:
            self._parse(self._xml)

    def _parse_radionuclide(self):
        """Parse radionuclide data."""
        for member in self._xml.findall(wfs.WFS_MEMBER):
            self._parse(member)

    def _parse_location_metadata(self, xml):
        """Parse location metadata."""
        for point in xml.findall(wfs.GML_POINT):
            fmisid = int(point.attrib[wfs.GML_ID].split('-')[-1])
            name = point.findtext(wfs.GML_NAME)
            location = tuple(float(p) for p in point.findtext(wfs.GML_POS).split())
            self.location_metadata[name] = dict({"fmisid": fmisid,
                                                 "latitude": location[0],
                                                 "longitude": location[1]
                                                 })
            self._location2name[location] = name

    def _parse(self, xml):
        """Parse data."""
        self._parse_location_metadata(xml)

        type2obs = _parse_names_and_units(xml)
        try:
            positions = _parse_positions(xml)
        except TypeError:
            print("No observations found")
            return
        latitudes = positions[::3]
        longitudes = positions[1::3]
        times = _parse_times(xml, positions)
        measurements = _parse_measurements(xml, (len(times), len(type2obs)))

        if self._timeseries:
            self._collect_timeseries(type2obs, latitudes, longitudes, times, measurements)
        else:
            self._collect_non_timeseries(type2obs, latitudes, longitudes, times, measurements)

    def _collect_timeseries(self, type2obs, latitudes, longitudes, times, measurements):
        for i, tim in enumerate(times):
            loc = (latitudes[i], longitudes[i])
            name = self._location2name[loc]
            if name not in self.data:
                self.data[name] = dict(times=[])
            self.data[name]["times"].append(tim)
            for j, key in enumerate(type2obs.keys()):
                if type2obs[key]["name"] not in self.data[name]:
                    self.data[name][type2obs[key]["name"]] = {"values": [], "unit": type2obs[key]["units"]}
                self.data[name][type2obs[key]["name"]]["values"].append(measurements[i, j])

    def _collect_non_timeseries(self, type2obs, latitudes, longitudes, times, measurements):
        for i, tim in enumerate(times):
            if tim not in self.data:
                self.data[tim] = dict()
            loc = (latitudes[i], longitudes[i])
            name = self._location2name[loc]
            if name not in self.data[tim]:
                self.data[tim][name] = dict()
            for j, key in enumerate(type2obs.keys()):
                self.data[tim][name][type2obs[key]["name"]] = dict({"value": measurements[i, j],
                                                                    "units": type2obs[key]["units"]
                                                                    })


def _parse_positions(xml):
    return np.fromstring(xml.findtext(wfs.GMLCOV_POSITIONS), dtype=float, sep=" ")


def _parse_times(xml, positions):
    times = np.array([dt.datetime(1970, 1, 1) + dt.timedelta(seconds=t) for t in positions[2::3]])
    if times.size == 0:
        times = np.array([dt.datetime.strptime(xml.findtext(wfs.GML_TIME_POSITION), TIME_FORMAT)])
    return times


def _parse_measurements(xml, shape):
    measurements = np.fromstring(xml.findtext(wfs.GML_DOUBLE_OR_NIL_REASON_TUPLE_LIST), dtype=float, sep=" ")
    return np.reshape(measurements, shape)


def _parse_names_and_units(xml):
    type2obs = dict()

    for field in xml.findall(wfs.SWE_FIELD):
        typ = field.attrib["name"]
        try:
            url = field.attrib[wfs.LINK]
            root = ET.fromstring(read_url(url))
            name = root.findtext(wfs.OMOP_LABEL)
            try:
                units = root.find(wfs.OMOP_UOM).attrib["uom"]
            except AttributeError:
                units = ''
        except KeyError:
            name = field.findtext(wfs.SWE_LABEL)
            units = field.find(wfs.SWE_UOM).attrib['code']
        type2obs[typ] = dict({"name": name, "units": units})

    return type2obs


def download_and_parse(query_id, args=None):
    """Download and parse the given stored query."""
    timeseries = False
    if args is None:
        args = []
    if "timeseries=True" in args:
        timeseries = True
        args.remove("timeseries=True")
    url = wfs.STORED_QUERY_URL + query_id
    if args:
        url = url + "&" + "&".join(args)
    xml = read_url(url)
    return MultiPoint(xml, query_id, timeseries=timeseries)



# -*- coding: utf-8 -*-
#
# Copyright (c) Panu Lahtinen
#
# Author(s):
#
#   Panu Lahtinen <pnuu+git@iki.fi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import defusedxml.ElementTree as ET
import datetime as dt

import numpy as np

from fmiopendata import wfs
from fmiopendata.utils import read_url

TIME_FORMAT = "%Y-%m-%dT%H:%M:%SZ"


class Lightning(object):
    """Class for holding lightning data."""

    def __init__(self, xml, mode):
        """Initialize the class."""
        self._xml = ET.fromstring(xml)
        self.latitudes = None
        self.longitudes = None
        self.times = None
        self.multiplicity = None
        self.peak_current = None
        self.cloud_indicator = None
        self.ellipse_major = None
        if mode == "simple":
            self._parse_simple()
        elif mode == "multipointcoverage":
            self._parse_multipoint()
        else:
            raise NotImplementedError("No parser for %s" % mode)

    def _parse_simple(self):
        """Parse lightning data.

        Version for fmi::observations::lightning::simple query.

        """
        lats, lons, times, flash_ids = [], [], [], []
        multiplicity, peak_current, cloud_indicator, ellipse_major = [], [], [], []
        for member in self._xml.findall(wfs.WFS_MEMBER):
            flash_id = _get_flash_id(member)
            tim = dt.datetime.strptime(member.findtext(wfs.WFS_TIME), TIME_FORMAT)
            lat, lon = [float(p) for p in member.findtext(wfs.GML_POS).split()]
            if flash_id not in flash_ids:
                lats.append(lat)
                lons.append(lon)
                times.append(tim)
                flash_ids.append(flash_id)
            param = member.findtext(wfs.WFS_PARAMETER_NAME)
            val = member.findtext(wfs.WFS_PARAMETER_VALUE)
            if param == "multiplicity":
                multiplicity.append(int(val))
            elif param == "peak_current":
                peak_current.append(float(val))
            elif param == "cloud_indicator":
                cloud_indicator.append(int(val))
            elif param == "ellipse_major":
                ellipse_major.append(float(val))
        self.latitudes = np.array(lats)
        self.longitudes = np.array(lons)
        self.times = np.array(times)
        self.multiplicity = np.array(multiplicity).astype(np.uint8)
        self.peak_current = np.array(peak_current)
        self.cloud_indicator = np.array(cloud_indicator).astype(np.uint8)
        self.ellipse_major = np.array(ellipse_major)

    def _parse_multipoint(self):
        """Parse lightning data.

        Version for fmi::observations::lightning::multipointcoverage query.

        """
        try:
            positions = np.fromstring(self._xml.findtext(wfs.GMLCOV_POSITIONS), dtype=float, sep=" ")
        except TypeError:
            print("No observations found")
            self._set_empty_observations()
            return
        self.latitudes = positions[::3]
        self.longitudes = positions[1::3]
        times = positions[2::3]
        self.times = np.array([dt.datetime.utcfromtimestamp(t) for t in times])

        data = np.fromstring(self._xml.findtext(wfs.GML_DOUBLE_OR_NIL_REASON_TUPLE_LIST), dtype=float, sep=" ")
        fields = [f.attrib['name'] for f in self._xml.findall(wfs.SWE_FIELD)]
        for i, field in enumerate(fields):
            vals = data[i::len(fields)]
            setattr(self, field, vals)

    def _set_empty_observations(self):
        self.latitudes = np.array([])
        self.longitudes = np.array([])
        self.times = np.array([])
        self.multiplicity = np.array([], dtype=np.uint8)
        self.peak_current = np.array([])
        self.cloud_indicator = np.array([], dtype=np.uint8)
        self.ellipse_major = np.array([])


def _get_flash_id(member):
    # <BsWfs:BsWfsElement gml:id="BsWfsElement.921.1">
    bs_wfs_element = member.find(wfs.WFS_BS_WFS_ELEMENT)
    # "BsWfsElement.921.1"
    full_id = bs_wfs_element.attrib[wfs.GML_ID]
    # 921
    flash_id = int(full_id.split(".")[1])

    return flash_id


def download_and_parse(query_id, args=None):
    """Download and parse the given stored query."""
    url = wfs.STORED_QUERY_URL + query_id
    if args:
        url = url + "&" + "&".join(args)
    xml = read_url(url)
    mode = query_id.split("::")[-1]
    return Lightning(xml, mode)



# -*- coding: utf-8 -*-
#
# Copyright (c) Panu Lahtinen
#
# Author(s):
#
#   Panu Lahtinen <pnuu+git@iki.fi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime as dt
import defusedxml.ElementTree as ET
import tempfile

import rasterio
import numpy as np

from fmiopendata import wfs
from fmiopendata.utils import read_url

TIME_FORMAT = "%Y-%m-%dT%H:%M:%SZ"

meta_cache = dict()


def get_meta(meta_url):
    """Get metadata from *meta_url*."""
    meta = meta_cache.get(meta_url)
    if meta is None:
        meta = ET.fromstring(read_url(meta_url))
        meta_cache[meta_url] = meta
    return meta


class Radar(object):
    """Container for Radar data."""

    def __init__(self):
        """Initialize class."""
        self.time = None
        self._gain = None
        self._offset = None
        self.elevation = None
        self.etop_threshold = None
        self.projection = None
        self.max_velocity = None
        self.url = None
        self.data = None
        self._dtype = None
        self.name = None
        self.label = None
        self.unit = None

    def download(self):
        """Download the data."""
        if self.data is None:
            with tempfile.NamedTemporaryFile() as fid:
                data = read_url(self.url)
                if len(data) < 10e3 and "ServiceException" in str(data):
                    msg = "WMS returned an exception: %s" % str(data)
                    raise ValueError(msg)
                fid.write(data)
                img = rasterio.open(fid.name)
                self.projection = img.crs.wkt
                self.data = img.read()
                self._dtype = self.data.dtype

    def get_area_mask(self):
        """Get a mask for areas outside the detection range."""
        self.download()
        if self.data.dtype == np.uint8:
            value = 255
        elif self.data.dtype == np.uint16:
            value = 65535
        else:
            if self._dtype == np.uint8:
                max_val = 255
            else:
                max_val = 65535
            value = max_val * self._gain + self._offset
        return self.data == value

    def get_data_mask(self):
        """Get a mask for invalid data."""
        self.download()
        if self.data.dtype in (np.uint8, np.uint16):
            value = 0
        else:
            value = 0 * self._gain + self._offset
        return self.data == value

    def calibrate(self):
        """Calibrate the data."""
        self.download()
        if self._gain:
            self.data = self.data * self._gain
            self.data += self._offset


class ParseRadar(object):
    """Parse radar data."""

    def __init__(self, xml):
        """Initialize class."""
        self._xml = ET.fromstring(xml)
        self.data = []
        self.times = []
        self._parse()

    def _parse(self):
        """Parse XML."""
        for member in self._xml.findall(wfs.WFS_MEMBER):
            radar = Radar()
            times = member.findall(wfs.GML_TIME_INSTANT)
            tim = dt.datetime.strptime(times[0].findtext(wfs.GML_TIME_POSITION),
                                       TIME_FORMAT)
            radar.time = tim
            self.times.append(tim)
            for parameter in member.findall(wfs.OM_PARAMETER):
                val = float(parameter.findtext(wfs.GML_MEASURE))
                name = parameter.find(wfs.OM_NAME).attrib[wfs.LINK]
                if "linearTransformationGain" in name:
                    radar._gain = val
                elif "linearTransformationOffset" in name:
                    radar._offset = val
                elif "reflectivityTreshold" in name:
                    radar.etop_threshold = val
                elif "elevationAngle" in name:
                    radar.elevation = val
                elif "maxVel" in name:
                    radar.max_velocity = val
            radar.name = member.find(wfs.SWE_DATA_RECORD).find(wfs.SWE_FIELD).attrib["name"]
            meta_url = member.find(wfs.SWE_DATA_RECORD).find(wfs.SWE_FIELD).attrib[wfs.LINK]
            meta = get_meta(meta_url)
            radar.unit = meta.find(wfs.OMOP_UOM).attrib["uom"]
            radar.label = meta.findtext(wfs.OMOP_LABEL)
            radar.url = member.findtext(wfs.GML_FILE_REFERENCE)
            radar.projection = radar.url.split('srs=')[-1].split('&')[0]
            self.data.append(radar)


def download_and_parse(query_id, args=None):
    """Download and parse the given stored query."""
    url = wfs.STORED_QUERY_URL + query_id
    if args:
        url = url + "&" + "&".join(args)
    xml = read_url(url)
    return ParseRadar(xml)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) Panu Lahtinen
#
# Author(s):
#
#   Panu Lahtinen <pnuu+git@iki.fi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
import datetime as dt

from fmiopendata.wms import get_wms_layers

TIME_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"


def write_title(fid, layer):
    """Write title of the layer."""
    fid.write("<h2>%s</h2>" % layer.title)


def write_abstract(fid, layer):
    """Write abstract of the layer."""
    if layer.abstract:
        fid.write("<p>%s</p>" % layer.abstract)


def write_name(fid, layer):
    """Write name of the layer."""
    fid.write("<li>Layer ID: %s</li>" % layer.name)


def write_bbox(fid, layer):
    """Write bounding boxes and CRSs available for the layer."""
    if layer.bbox:
        fid.write("<li>Bounding boxes:</li>")
        fid.write("<ul>")
        bboxes = layer.bbox
        for bbox in bboxes:
            fid.write("<li>%s</li>" % bbox['CRS'])
            fid.write("<ul>")
            fid.write("<li>X min: %s</li>" % bbox['minx'])
            fid.write("<li>X max: %s</li>" % bbox['maxx'])
            fid.write("<li>Y min: %s</li>" % bbox['miny'])
            fid.write("<li>Y max: %s</li>" % bbox['maxy'])
            fid.write("</ul>")
        fid.write("</ul>")


def write_times(fid, layer):
    """Write earliest and latest times, and time step for the layer."""
    if layer.time_step_str:
        fid.write("<li>Available times:</li>")
        fid.write("<ul>")
        fid.write("<li>Earliest: %s</li>" % dt.datetime.strftime(min(layer.times), TIME_FORMAT))
        fid.write("<li>Latest: %s</li>" % dt.datetime.strftime(max(layer.times), TIME_FORMAT))
        fid.write("<li>%s</li>" % layer.time_step_str)
        fid.write("</ul>")


def write_elevations(fid, layer):
    """Write available elevations for the layer."""
    if layer.elevations:
        fid.write("<li>Elevations:</li>")
        fid.write("<ul>")
        for elev in layer.elevations:
            fid.write("<li>%s</li>" % elev)
        fid.write("</ul>")


def write_html(fname, layers):
    """Save HTML page showing the WMS layers."""
    with open(fname, 'w') as fid:
        fid.write("<html><body>")
        fid.write("<h1>Available WMS layers in FMI open data.</hi>")
        for key in sorted(layers):
            layer = layers[key]
            write_title(fid, layer)
            write_abstract(fid, layer)
            fid.write("<ul>")
            write_name(fid, layer)
            write_bbox(fid, layer)
            write_times(fid, layer)
            write_elevations(fid, layer)
            fid.write("</ul>")
        fid.write("</body></html>")


def main():
    """Run the script."""
    try:
        fname = sys.argv[1]
    except IndexError:
        fname = "wms.html"

    layers = get_wms_layers()
    write_html(fname, layers)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) Panu Lahtinen
#
# Author(s):
#
#   Panu Lahtinen <pnuu+git@iki.fi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import defusedxml.ElementTree as ET

import sys

from fmiopendata import wfs
from fmiopendata.utils import read_url


def write_title(fid, query):
    """Write title of the query."""
    fid.write("<h2>%s</h2>" % query["title"])


def write_description(fid, query_id):
    """Write available query parameters."""
    xml = ET.fromstring(read_url(wfs.BASE_URL + "DescribeStoredQueries&storedquery_id=" + query_id))
    fid.write("<p>")
    fid.write(xml.findtext(wfs.WFS_ABSTRACT).strip())
    fid.write("</p>")
    fid.write("<ul>")
    fid.write("<li>Query ID: %s</li>" % query_id)
    fid.write("<li>Available arguments:</li>")
    fid.write("<ul>")
    params = xml.findall(wfs.WFS_PARAMETER)
    for i, param in enumerate(params):
        fid.write("<li>%s</li>" % param.attrib["name"])
        fid.write("<ul>")
        param_title = param.findtext(wfs.WFS_TITLE)
        fid.write("<li>%s</li>" % param_title)
        param_abstract = param.findtext(wfs.WFS_ABSTRACT).strip()
        fid.write("<li>%s</li>" % param_abstract)
        fid.write("</ul>")
    fid.write("</ul>")
    fid.write("</ul>")
    fid.write("</ul>")


def write_html(fname, queries):
    """Save HTML page showing the WFS stored queries."""
    with open(fname, 'w') as fid:
        fid.write("<html><body>")
        fid.write("<h1>Available WFS stored queries in FMI open data.</hi>")
        for key in sorted(queries):
            query = queries[key]
            write_title(fid, query)
            write_description(fid, key)
        fid.write("</body></html>")


def write_title_md(fid, query):
    """Write title of the query."""
    fid.write("## %s\n\n" % query["title"])


def write_md(fname, queries):
    """Save markdown page showing the WFS stored queries in markdown."""
    with open(fname, 'w') as fid:
        fid.write("# Available WFS stored queries in FMI open data.\n\n")
        for key in sorted(queries):
            query = queries[key]
            write_title_md(fid, query)
            write_description_md(fid, key)


def write_description_md(fid, query_id):
    """Write available query parameters in markdown."""
    xml = ET.fromstring(read_url(wfs.BASE_URL + "DescribeStoredQueries&storedquery_id=" + query_id))
    fid.write(xml.findtext(wfs.WFS_ABSTRACT).strip())
    fid.write("\n\n")
    fid.write("* Query ID: `%s`\n" % query_id)
    fid.write("* Available arguments:\n")
    params = xml.findall(wfs.WFS_PARAMETER)
    for i, param in enumerate(params):
        fid.write("    * %s\n" % param.attrib["name"])
        param_title = param.findtext(wfs.WFS_TITLE)
        fid.write("        * %s\n" % param_title)
        param_abstract = param.findtext(wfs.WFS_ABSTRACT).strip()
        fid.write("        * %s\n" % param_abstract)
    fid.write("\n\n")


def main():
    """Run the script."""
    try:
        fname = sys.argv[1]
    except IndexError:
        fname = "wfs.html"

    queries = wfs.get_stored_queries()
    if fname.endswith("html"):
        write_html(fname, queries)
    else:
        write_md(fname, queries)


if __name__ == "__main__":
    main()
